from infi.shell.commands import commandgroup, command, render
from infi.shell.commands.types import Str, Int, Bool, IPAddress, SubnetMask
from infi.shell.exceptions import BadValue, CommandAborted, NoActiveReadlineContext
from infi.shell.rlcontext import get_active_rlcontext
from infi.shell.config import config

from .types import ConfigParameter
from . import superuser

from bunch import Bunch

def _convert_boolean(str_value):
    str_value = str_value.lower().strip()
    if str_value in ("true", "1", "yes"):
        return True
    if str_value in ("false", "0", "no"):
        return False
    raise BadValue()

@commandgroup(prefix="config", provides=(ConfigParameter,))
class ConfigCommands(object):
    _BASE_URI = "config"
    _TYPES = {key : (value, converter)
              for value, keys, converter in (
                  ("Number",   (int, long, float),          None),
                  ("Text",     (basestring, str, unicode),  None),
                  ("Boolean",  (bool,),                     _convert_boolean),
                  ("List",     (list, tuple, set),          None),
                  )
              for key in keys}
    # Definition of validations of configuration values. When setting
    # configuration values using the 'config.set' command, the validation
    # method which corresponds with the parameter's name is invoked. The method
    # should return either None if the value is valid, or an error string if
    # the value is invalid.
    # It's important to keep in mind that the purpose of this validation code
    # is to work around problems with the management configuration API.
    # Generally, InfiniShell should contain as little application logic as
    # possible, and should not be coupled with management code.
    _VALIDATION_PRESETS = Bunch(
            NOT_EMPTY = lambda value: None if value else "Empty value not allowed",
            )
    _VALIDATIONS = {
            "mgmt/environment.dns_servers" : _VALIDATION_PRESETS.NOT_EMPTY,
            }
    @superuser
    @command(types=(Str,), substitute=dict(category_name="category"))
    @render(order="parameter type value".split(), alignment="left")
    def query(self, _api, category_name=None):
        if category_name is not None:
            category = _api.get("{}/{}".format(self._BASE_URI, category_name))
            return self._flatten_category({ category_name : category })
        categories = _api.get(self._BASE_URI)
        parameters = []
        for category_name, category in categories.iteritems():
            parameters.extend(self._flatten_category({ category_name : category }))
        parameters.sort(key=lambda p: p.parameter)
        return parameters
    def _flatten_category(self, categories):
        for name, value in categories.iteritems():
            if isinstance(value, dict):
                for parameter in self._flatten_category(value):
                    parameter.parameter = "{}/{}".format(name, parameter.parameter)
                    yield parameter
            else:
                parameter = Bunch(
                        parameter=name,
                        value=value,
                        )
                yield parameter
    @superuser
    @command(types=(ConfigParameter, Str))
    def set(self, _api, parameter, value):
        self._validate_parameter_value(parameter, value)
        old_value = _api.get("{}/{}".format(self._BASE_URI, parameter))
        if not isinstance(old_value, basestring):
            value_type = type(old_value)
            assert value_type in self._TYPES, "Unsupported parameter type: " + value_type
            value_type_name, converter = self._TYPES[value_type]
            converter = converter or value_type
            try:
                value = converter(value)
            except:
                raise BadValue('Cannot convert "{}" to {}'.format(value, value_type_name.lower()))
        _api.put("{}/{}".format(self._BASE_URI, parameter), value)
        return 'Changed configuration value "{}" from "{}" to "{}"'.format(parameter, old_value, value)
    def _validate_parameter_value(self, parameter, value):
        validator = self._VALIDATIONS.get(parameter)
        if validator:
            error = validator(value)
            if error is not None:
                raise BadValue('Cannot set "{}" to "{}": {}'.format(parameter, value, error))
    def _configure_smtp(self, _api, config_name, host, port, tls, username, password, fromaddress):
        _api.put("{}/{}".format(self._BASE_URI, config_name), dict(
            host=host,
            port=port,
            tls=tls,
            username=username,
            password=password,
            fromAddress=fromaddress,
            ))
    def _test_smtp(self, _api, config_name, recipient):
        _api.post("{}/{}/test".format(self._BASE_URI, config_name), recipient)
        return 'Test email sent to {}'.format(recipient)
    @command(
        description="Display SMTP client settings"
        )
    def smtp_query(self, _api):
        return self.query(_api, "smtp")
    @command(
        description="Set SMTP client settings",
        types=(Str, Int, Bool, Str, Str, Str),
        )
    def smtp_set(self, _api, host, port, tls, username, password, fromaddress):
        self._configure_smtp(_api, "smtp", host, port, tls, username, password, fromaddress)
        return 'SMTP configuration updated'
    @command(
        description="Test current SMTP client settings",
        types=(Str,)
        )
    def smtp_test(self, _api, recipient):
        return self._test_smtp(_api, "smtp", recipient)
    @superuser
    @command
    def infinidat_smtp_query(self, _api):
        return self.query(_api, "infinidat_smtp")
    @superuser
    @command(types=(Str, Int, Bool, Str, Str, Str))
    def infinidat_smtp_set(self, _api, host, port, tls, username, password, fromaddress):
        self._configure_smtp(_api, "infinidat_smtp", host, port, tls, username, password, fromaddress)
        return 'Infinidat SMTP configuration updated'
    @superuser
    @command(types=(Str,))
    def infinidat_smtp_test(self, _api, recipient):
        return self._test_smtp(_api, "infinidat_smtp", recipient)
    @command(
        description="Display system's IP address",
        )
    @render(individually=True)
    def system_ip_query(self, _api):
        return _api.get("{}/ip_config".format(self._BASE_URI))
    @command(
        description="Set system's IP address",
        types=(IPAddress, SubnetMask, IPAddress),
        )
    def system_ip_set(self, _api, ip, netmask, gateway):
        if not config.auto_confirm_prompts:
            try:
                rlcontext = get_active_rlcontext()
            except NoActiveReadlineContext:
                # This exception should be part of Infinidat common shell code
                raise CommandAborted("Command requires user confirmation (use -y or the interactive shell)")
            current_config = self.system_ip_query(_api)
            template = (
                "IP       {ip}\n"
                "Netmask  {netmask}\n"
                "Gateway  {gateway}"
                )
            prompt = (
                "Current system IP configuration:\n"
                "{}\n\n"
                "New system IP configuration:\n"
                "{}\n\n"
                "Are you sure the new system IP configuration is correct?"
                .format(
                    template.format(**current_config),
                    template.format(ip=ip, netmask=netmask, gateway=gateway),
                    )
                )
            if not rlcontext.yesno(prompt):
                raise CommandAborted
        _api.put("{}/ip_config".format(self._BASE_URI), dict(
            ip=ip,
            netmask=netmask,
            gateway=gateway,
            ))
        return "System IP configuration updated"
